# -*- coding: UTF-8 -*-
###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# libsecure.py
#
# classes utilitaires pour lancement des services en https
#
###########################################################################

import httplib, xmlrpclib
from OpenSSL import SSL
from os.path import join, dirname, isfile, isdir
from os import unlink, system, makedirs
import sys
from ConfigParser import ConfigParser
from creole.parsedico import parse_dico
from twisted.python import log

# transport sécurisé utilisant un certificat
class TransportEole(xmlrpclib.SafeTransport):

    def __init__(self):
        dico = parse_dico()
        if dico.has_key('sso_cert') and isfile(dico['sso_cert']):
            self.certfile = dico['sso_cert']
        else:
            self.certfile = join('/usr/share/sso/certs/server.pem')
        super(xmlrpclib.SafeTransport, self).__init__()

    def make_connection(self, host):
        # create a HTTPS connection object from a host descriptor
        # host may be a string, or a (host, x509-dict) tuple
        localhost, extra_headers, x509 = self.get_host_info(host)
        try:
            HTTPS = httplib.HTTPS
        except AttributeError:
            raise NotImplementedError(
                "your version of httplib doesn't support HTTPS"
                )
        else:
            cx = HTTPS(localhost, None,
                       key_file = self.certfile,
                       cert_file = self.certfile)
            return cx

class ServerContextFactory:
    """
    Factory permettant de créer un contexte SSL
    """

    def __init__(self,certfile=''):
        if certfile != '':
            self.certfile = certfile
        else:
            self.certfile = join('/usr/share/sso/certs/server.pem')

    def getContext(self):
        """Create an SSL context.
        This is a sample implementation that loads a certificate from a file 
        called 'server.pem'."""
        ctx = SSL.Context(SSL.SSLv23_METHOD)
        ctx.use_certificate_file(self.certfile)
        ctx.use_privatekey_file(self.certfile)
        return ctx

def gen_certif(certfile):
    """génère un certificat pour le server https
    """
    conf_file = '/etc/sysconfig/eole/certif-eole.conf'
    if not isfile(conf_file):
        conf_file = '/usr/share/eole/certif-eole.conf'
        if not isfile(conf_file):
            conf_file = '/usr/share/eole/requete-eole.conf'
            if not isfile(conf_file):
                print '! pas de fichier de configuration trouvé pour la génération du certificat !'
                print '  vous pouvez fournir un certificat manuellement (%s)' % certfile
                print '  (inclure le certificat et la clé au format PEM)'
                sys.exit(1)
    requete_dir = dirname(certfile)
    if not isdir (requete_dir):
        makedirs(requete_dir)
    keyfile = "%s/server.key" % requete_dir
    tmp_keyfile = "/tmp/key"

    # génération certificat x509
    ssl_cmd = "openssl req -x509 -new -newkey rsa:1024 -days 1825 -config %s -keyout %s -out %s >/dev/null 2>&1" % (conf_file, tmp_keyfile, certfile)
    res = system(ssl_cmd)
    if res != 0:
        print('\n! erreur lors de la génération du certificat ssl dans %s !\n' % certfile)
        sys.exit(1)

    # génération de la clé publique
    key_cmd = "openssl rsa -in %s -passin pass:secret -out %s >/dev/null 2>&1" % (tmp_keyfile, keyfile)
    res = system(key_cmd)
    if res != 0:
        print('\n! erreur lors de la génération de la clé ssl dans %s !\n' % keyfile)
    unlink(tmp_keyfile)
    # finalisation du fichier certificat
    dico = parse_dico()
    cert_data = open(certfile).read()
    key_data = open(keyfile).read()
    f_cert = open(certfile,'w')
    f_cert.write("SubjectAltName=%s.%s\n%s\n%s" % (dico['nom_machine'], dico['nom_domaine_local'],cert_data.rstrip(),key_data.rstrip()))
    f_cert.close()

