# -*- coding: UTF-8 -*-
###########################################################################
# Eole NG - 2007  
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr 
#  
# util.py
#  
# fonctions utilitaires pour la librairie sso eole
# (actuellement seulement annuaire LDAP)
#       
###########################################################################
import urlparse, socket, re, IPy
from twisted.web2.http_headers import parseCookie
from twisted.python import log

def urljoin(url, params):
    """ajoute des paramètres à une url
    >>> urljoin('http://test/','a=b')
    'http://test/?a=b'
    >>> urljoin('http://test/?a=b','c=d')
    'http://test/?a=b&c=d'
    """
    urllist = list(urlparse.urlsplit(url))
    if urllist[-2] == '':
        urllist[-2] = params
    else:
        urllist[-2] += '&' + params
    return urlparse.urlunsplit(urllist)

head = """<head>
<meta http-equiv="Pragma" content="no-cache">
<meta http-equiv="Cache-Control" content="no-store">
<meta http-equiv='Expires' content='Mon, 26 Jul 1997 05:00:00 GMT'>
<title>Authentification</title>
<link rel='stylesheet' type='text/css' href='/css/leaves.css'>
</head>
"""

def unescape(url, quote=None):
    """supprime les caractères escapés dans une url"""
    entities = [('&gt;','>'), ('&lt;','<'), ('&amp;','&')]
    if quote:
        entities.insert(0, ('&quot;','"'))
    for substr, car in entities:
        url = url.replace(substr, car)
    return url

def get_service_from_url(url):
    """supprime la partie requête d'une url pour permettre d'indentifier le service
    """
    url = unescape(url)
    u = urlparse.urlparse(url)
    service = "%s://%s%s" % (u.scheme, u.netloc, u.path)
    return service

def getCookies(ctx):
    """renvoie la liste  des cookies sous forme Twisted"""
    cookiestr = ctx.headers.getRawHeaders('Cookie')
    if cookiestr:
        return parseCookie(cookiestr)
    return []

def getCookie(ctx, key):
    """renvoie un cookie Twisted"""
    for cookie in getCookies(ctx):
        if cookie.name == key:
            return cookie
    return None

def verif_cert(conn, cert, err_num, err_depth, rc):
    """fonction de vérification du certificat"""
    return not cert.has_expired()

def check_hostname_by_ip(url, addr):
    """vérification de la validité de l'hôte de l'url en fonction d'une IP
    """
    try:
        # vérification de l'adresse
        ip = IPy.IP(addr)
        try:
            hostname = socket.gethostbyname(url.hostname)
        except:
            hostname = url.hostname
        assert ip.overlaps(hostname)
        return True
    except:
        return False
    return False

def check_hostname_by_domain(url, addr):
    """vérification de la validité de l'hôte de l'url par expression régulière
    """
    hostname = url.hostname
    # on essaye un match de l'expression régulière sur le nom d'hôte
    try:
        match_obj = re.match(addr, hostname)
    except:
        log.msg("""Erreur lors du test sur l'expression régulière '%s' (syntaxe invalide ?)""" % addr)
        return False
    return match_obj is not None
