#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authform.py
#
# service web de connexion au service sso Eole
# 
# compatible avec les versions 1 et 2 de la librairie CAS
# (http://www.ja-sig.org/products/cas/), a l'exception de la gestion
# des tickets de proxy (PGT)
#
###########################################################################

from socket import error as socket_err
import sys, xmlrpclib
from time import time
from twisted.internet import reactor
from twisted.web2 import static, server, channel, http, responsecode
from twisted.web2.resource import PostableResource as Resource
from twisted.web2.http_headers import Cookie as TwCookie
from twisted.web2.xmlrpc import XMLRPC
from config import *
from eolesso.util import *
from eolesso.libsecure import ServerContextFactory, gen_certif
from eolesso.dataproxy import LDAPProxy
from eolesso import ticketcache
from oidprovider import OIDRoot
from authserver import SSOSessionManager, log

try:
    _ = _
except NameError:
    _ = unicode

class XMLRPCManager(XMLRPC):

    def __init__(self,session_manager):
        self.manager = session_manager
        XMLRPC.__init__(self)

    def xmlrpc_authenticate(self, post_data, username, password):
        log.msg('* xmlrpc authenticate')
        return self.manager.authenticate(username, password)

    def xmlrpc_validate_session(self, post_data, session_id):
        log.msg('* xmlrpc validate_session')
        return self.manager.validate_session(session_id)

    def xmlrpc_verify_session_id(self, post_data, session_id):
        log.msg('* xmlrpc verify_session_id')
        return self.manager.verify_session_id(session_id)

    def xmlrpc_logout(self, post_data, session_id):
        log.msg('* xmlrpc logout')
        return self.manager.logout(session_id)

    def xmlrpc_verify_app_ticket(self, post_data, app_ticket, appurl):
        log.msg('* xmlrpc verify_app_ticket')
        return self.manager.verify_app_ticket(app_ticket, appurl)

    def xmlrpc_get_user_details(self, post_data, app_ticket, appurl, sections=False):
        log.msg('* xmlrpc get_user_details')
        return self.manager.get_user_details(app_ticket, appurl, sections)


class AuthenticationVerify(Resource):
    """ Page de vérification de l'identité 
        démarre une session et stocke un cookie dans le client
        si les identifiants sont corrects"""
    isLeaf = False
    addSlash = False

    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def render(self, request):
        try:
            username = unicode(request.args['username'][0],encoding)
            password = unicode(request.args['password'][0],encoding)
        except:
            return http.Response(stream="<html>%s<body>Accès non autorisé !</body></html>"%head)
        try:
            from_url = request.args['service'][0]
        except:
            from_url = ""
        # vérification du login ticket
        try:
            pass
            #ticket = request.args['lt'][0]
            #assert self.manager.login_sessions.validate_session(ticket)
        except:
            # on n'est pas passé par /login : on renvoie dessus
            if from_url:
                response = http.RedirectResponse('/login?service=%s' % from_url)
            else:
                response = http.RedirectResponse('/login')
            return response
        # on va tenter l'authentification sur tous les serveurs d'auth 
        session_id = self.manager.authenticate(username, password)
        # authentification distante (SSO) reussie
        if session_id != "":
            # on stocke un cookie dans le navigateur
            # chaine = '%s,%s'%(session_id), username.encode(encoding))
            session_id = session_id.encode(encoding)
            cookies = getCookies(request)
            for cookie in cookies:
                if cookie.name == 'EoleSSOServer':
                    cookies.remove(cookie)
            cookies.append(TwCookie("EoleSSOServer", session_id, path = "/", discard=True, secure=True))
            if from_url == "":
                response = http.RedirectResponse('/loggedin')
            else:
                # on redirige sur la page de l'application de destination
                app_ticket = self.manager.get_app_ticket(session_id, from_url)
                response = http.RedirectResponse(urljoin(from_url, 'ticket=%s' % app_ticket))
                log.msg("redirecting to calling app with ticket : %s" % from_url)
            if cookies:
                response.headers.setHeader('Set-Cookie', cookies)
            return response
    
        # aucun des serveurs n'a authentifié l'utilisateur (meme l'authentification locale), on revient
        # sur la page de saisie de username/password
        return http.RedirectResponse('/?service=%s&failed=1' %(from_url))


class CASCompliantResponse(Resource):
    """implémentation du protocole CAS pour la validation de ticket
    """
    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def _gen_user(self, infos, data_filter):
        """génération de la partie données utilisateur de la réponse CAS
        """
        data = ""
        if data_filter != "":
            for section, attrs in data_filter.items():
                data += "    <cas:%s>\n%s    </cas:%s>\n" % (section, self._gen_section(attrs,infos), section)
        else:
            data = "    <cas:user>%s</cas:user>\n" % infos['uid'][0]
        return data

    def _gen_section(self,attrs,infos):
        """génère une section dans la réponse
        """
        section = ""
        for name, datakey in attrs.items():
            for val in infos[datakey]:
                section += "      <cas:%s>%s</cas:%s>\n" % (name, val, name)
        return section

    def render(self, request):
        error = {'code':'INTERNAL_ERROR','detail':''}
        ok = False
        try:
            from_url = request.args['service'][0]
            ticket = request.args['ticket'][0]
        except (IndexError, KeyError):
            error = {'code':'INVALID_REQUEST','detail':'paramètre manquant'}
            data = error, None
        else:
            #on va tenter l'authentification sur tous les serveurs d'auth 
            ok, data = self.manager.get_user_details(ticket, from_url, True)
            infos, filter_data = data
        if ok:
            #authentification distante (SSO) reussie
            # génération des informations utilisateur
            user_infos = self._gen_user(infos, filter_data)
            # formatting response
            if CAS_VERSION == 2:
                response = """<?xml version="1.0" encoding="UTF-8"?>
  <cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
    <cas:authenticationSuccess>
    %s    </cas:authenticationSuccess>
  </cas:serviceResponse>""" % user_infos
            else:
                response = 'yes\n%s\n' % infos['uid'][0]
        else:
            error, filter_data  = data
            # aucun des serveurs n'a authentifié l'utilisateur (meme l'authentification locale), on revient
            # sur la page de saisie de username/password
            if CAS_VERSION == 2:
                response = """<?xml version="1.0" encoding="UTF-8"?>
  <cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
    <cas:authenticationFailure code="%(code)s">
    %(detail)s
    </cas:authenticationFailure>
  </cas:serviceResponse>""" % error
            else:
                response = 'no\n\n'
        cas_resp = http.Response(stream=response)
        cas_resp.headers.setRawHeaders('Pragma', 'no-cache')
        cas_resp.headers.setRawHeaders('Cache-Control', 'no-store')
        cas_resp.headers.setRawHeaders('Expires', 'Mon, 26 Jul 1997 05:00:00 GMT')
        return cas_resp

class LoggedInForm(Resource):
    """ 
    page de confirmation d'ouverture de session (si pas d'url de redirection fournie)
    """
    addSlash = False

    def render(self,request):
        return http.Response(stream="""<html>%s<body><p><H1>Session SSO valide</H1></p>
        <p><form action="/logout" method="post">
        <input type="Submit" value="Fermer la connexion"></p></body></html>""" % head)

class ErrorPage(Resource):
    """ 
    page d'erreur http
    """
    addSlash = False

    def __init__(self, code, description):
        self.code = code
        self.description = description
        title = ('Eole-SSO : Erreur')
        super(ErrorPage, self).__init__()

    def render(self,request):
        return http.StatusResponse(code=self.code, description=self.description)

class LogoutForm(Resource):
    """ 
    page de confirmation de logout (si pas d'url de redirection fournie)
    """
    addSlash = False

    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def render(self,request):
        # invalidation de la session sur le serveur d'authentification
        cookies = getCookies(request)
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                session_id = cookie.value
                if session_id is not None:
                    # suppression de la session sur le manager
                    self.manager.logout(session_id.encode(encoding))
                    cookie.expires = 1
                break
        if request.args.has_key('service'):
            url_from = request.args['service'][0]
            response = http.RedirectResponse(url_from)
            log.msg("logged out : redirecting to : %s" % url_from)
        else:
            #si pas de parametre from, on affiche la page de logout en local
            response = http.Response(stream="""<html>%s<body><p><H1>Fin de la session SSO</H1></p></body></html>
            <p><form action="/login" method="post">
            <input type="Submit" value="Connexion"></p></body></html>""" % head)
        # mise en place des cookies
        if cookies:
            response.headers.setHeader('Set-Cookie', cookies)
        return response

class TestForm(Resource):
    """The resource that is returned when you are not logged in"""
    addSlash = False
    
    def __init__(self, manager, auth_server, oidserver, *args):
        self.manager = manager
        self.auth_server = auth_server
        self.oidserver = oidserver
        self.content = ''
        Resource.__init__(self)

    def is_true(self, param):
        return param == "true"
        
    def render_content(self, request):
        """
        protocole CAS:
         1. service : l'URL d'où on vient
         2. renew=true : on force à se réauthentifier
         3. gateway=true : si pas authentifié, on revient à l'URL
            définie par service sans ticket d'application
        """
        self.content = ""
        gateway = self.is_true(request.args.get('gateway', ['false'])[0])
        renew = self.is_true(request.args.get('renew', ['false'])[0])
        warn = self.is_true(request.args.get('warn', ['false'])[0])
        try:
            from_url = request.args['service'][0]
        except KeyError:
            # nécessaire car les ressources ne sont instanciées qu'une
            # seule fois
            from_url = ''
        #on va verifier si on a un identifiant de session dans le cookie
        user_session = getCookie(request, 'EoleSSOServer')
        # XXX: IE et Opera ne gerent pas les cookies de la meme
        # maniere que Firefox sous firefox, quand un cookie devient
        # invalide (au niveau du temps), il est supprimé sous IE et
        # Opera, ils sont gardés, on doit donc tester si ils ont une
        # valeur spéciale, ici foo
        #si on en a bien un, on le verifie auprès du serveur d'authentification
        if not renew and user_session and user_session != 'foo':
            session_id = user_session.value.encode(encoding)
            verif =  self.manager.verify_session_id(session_id)
            # si l'ID de session est verifie, on génère le ticket applicatif,
            # et on redirige vers l'URL de base
            if verif[0]:
                if from_url:
                    app_ticket = self.manager.get_app_ticket(session_id, from_url)
                    url = urljoin(from_url, 'ticket=%s' % app_ticket)
                    if warn:
                        self.content = """<html>%s<body onload="alert('%s'); document.location.href='%s'"></body></html>""" % (head, 'you are being authenticated', url)
                        return None
                    if from_url == '':
                        self.content = "<html>%s<body>pas de destination connue :)</body></html>"%head
                        return None
                    return url
                return '/loggedin'
            
        if gateway and from_url:
            return urljoin(from_url, 'ticket=')
        self.auth_msg = "Veuillez vous authentifier"
        if request.args.has_key('failed'):
            if request.args['failed'][0] == '1':
                self.auth_msg = "Echec de l'authentification, recommencez"
        # création d'un login ticket jouable une seule fois pour cet essai d'authentification
        login_ticket = self.manager.login_sessions.add_session('login')
        if from_url != "":
            verify_form = """<input type="hidden" name="service" value="%s"/>""" % (from_url)
        else:
            verify_form = ""

        self.content = """
        <html>
        %s
        <body>
        <div id="container">
           <div id="sitename">
              <h1>%s</h1>
           </div>
           <div id="wrap1">
           <div id="wrap2">
              <div id="content">
                 <form action='/verify/' method='post'>
                 <input type="hidden" name="lt" value="%s">
                    %s
                    <table>
                       <tr><td>Login : </td><td><input type='text' name='username'/></td></tr>
                       <tr><td>Mot de passe : </td><td><input type='password' name='password'/></td></tr>
                    </table>
                    <input type='submit'/>
                 </form>
              </div>
           </div>
           </div>
        </div>
        </body>
        </html>
        """ % (head, self.auth_msg, login_ticket, verify_form)
        return None

    def render(self,request):
        redirect_to = self.render_content(request)
        if redirect_to == None:
            return http.Response(stream=self.content)
        else:
            log.msg("authform, redirecting to : %s" % redirect_to)
            return http.RedirectResponse(redirect_to)

    def locateChild(self, request, segments):
        if segments and segments[0] == 'verify':
            return AuthenticationVerify(self.manager), ()
        if segments and segments[0] in ('validate','proxyValidate','serviceValidate'):
            return CASCompliantResponse(self.manager), ()
        if segments and segments[0] == 'loggedin':
            return LoggedInForm(), ()
        if segments and segments[0] == 'logout':
            return LogoutForm(self.manager), ()
        if segments and segments[0] == 'xmlrpc':
            return self.auth_server, ()
        if segments and segments[0] == 'oid':
            return self.oidserver, segments[1:]
        if segments and segments[0] == 'css':
            return static.File('interface'), segments[1:]
        if segments[0] == 'images':
            return static.File('interface/images'), segments[1:]
        if segments and segments[0] in ['login','']:
            return self, () 
        return ErrorPage(responsecode.NOT_FOUND, u"Ressource non trouvée %s" % '/'.join(segments)), ()

def run():
    if '--test' in sys.argv:
        import doctest
        doctest.testmod(verbose=True)
    else:
        global PARENT_URL
        # intialisation du serveur
        static_data={'rne': [RNE], 'nom_etab': [ETABLISSEMENT]}
        proxy = LDAPProxy(LDAP_SERVER, static_data)
        # lancement du serveur
        if PARENT_URL == '':
            PARENT_URL = None
        log.startLogging(sys.stdout)
        # manager de sessions
        manager = SSOSessionManager(proxy, PARENT_URL)
        # frontend xmlrpc
        auth_server = XMLRPCManager(manager)
        # serveur openid
        baseurl = "https://%s:%s/oid/" % (AUTH_SERVER_ADDR, str(EOLESSO_PORT))
        oidserver = OIDRoot(manager, baseurl)
        # serveur CAS (racine du site)
        root = TestForm(manager, auth_server, oidserver)
        # si le certificat n'existe pas, on le génère
        if not os.path.isfile(CERTFILE):
            log.msg("** Génération d'un nouveau certificat ssl ...")
            gen_certif(CERTFILE)
        log.msg("** Initialisation du serveur SSO")
        reactor.listenSSL(int(EOLESSO_PORT), channel.HTTPFactory(server.Site(root)),
                          ServerContextFactory(CERTFILE))
        reactor.run()

if __name__ == '__main__':
    run()
