#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authserver.py
#
# serveur xmlrpc de gestion de session SSO Eole
#
###########################################################################

import socket
import xmlrpclib
import ldap
from md5 import md5
from glob import glob
from ConfigParser import ConfigParser
# imports twisted
from twisted.python import log
# imports sso eole
from config import *
from eolesso.util import *
from eolesso import ticketcache
from eolesso.ticketcache import InvalidSession
# import des fonctions de calcul de données utilisateur
import user_infos



class SSOSessionManager(object):

    def __init__(self,data_proxy,parent=None):
        self.address = socket.getfqdn()
        # cache des "granting tickets (ticket de session sso)"
        self.user_sessions = ticketcache.TicketCache(SESSION_TIMER, self.address, 'TGC', renew_tickets=True)
        # cache des "application tickets"
        self.app_sessions = ticketcache.TicketCache(APP_TIMER, self.address, 'ST')
        # cache des "login tickets" (pour empêcher les navigateurs de 'rejouer' l'authentification avec un cache)
        self.login_sessions = ticketcache.TicketCache(APP_TIMER, self.address, 'LT')
        # proxy d'authentification
        self._data_proxy = data_proxy
        self._parent = None
        if parent:
            self._parent = xmlrpclib.ServerProxy(parent)
            log.msg("\nServeur parent déclaré : %s" % parent)
        self.filters = {}
        self.apps = {'all':{}}
        self.secure_apps = {'all':{}}
        self.load_filters()


    def load_applications(self, config):
        """Chargement des applications associées aux filtres"""
        loaded_apps = []
        for app in config.sections():
            try:
                app_infos = dict(config.items(app))
                try:
                    port = app_infos['port'] = int(app_infos['port'])
                except:
                    port = app_infos['port'] = 'all'
                scheme = app_infos['scheme']
                if scheme == 'https':
                    app_stores = [self.secure_apps]
                elif scheme == 'http':
                    app_stores = [self.apps]
                else:
                    app_stores = [self.apps, self.secure_apps]
                # on trie les applications par port et chemin
                for store in app_stores:
                    port_store = store.setdefault(port, {})
                    port_store[app_infos['baseurl']] = [scheme, 
                                                        app_infos['addr'],
                                                        app_infos['typeaddr'],
                                                        app_infos['filter']]
                    loaded_apps.append(app)
            except Exception, e:
                log.msg("! Error loading application %s : %s !" % (app, e))
        return loaded_apps
    
   
    def load_filters(self):
        """chargement des filtres de données et applications
        """
        loaded_apps = []
        for conf_file in glob('%s/*.ini' % FILTER_DIR):
            conf_filename = os.path.splitext(os.path.basename(conf_file))[0]
            config = ConfigParser()
            config.read(conf_file)
            if conf_filename in ['applications', 'local_apps']:
                # liste des applications associées aux filtres
                loaded_apps += self.load_applications(config)
            else:
                # ajout d'un filtre de données
                app_filter = dict( (section, dict(config.items(section)))
                                   for section in config.sections() )
                self.filters[conf_filename] = app_filter
                log.msg("- Filtre chargé: %s" % conf_filename)
        log.msg("- Applications chargées: %s" % ", ".join(loaded_apps))


    def authenticate(self, username, password):
        sessionID = ''
        if type(username) == unicode: username = username.encode(encoding)
        if type(password) == unicode: password = password.encode(encoding)

        # on demande en premier au serveur père qu'il vérifie les informations
        if self._parent is not None:
            sessionID = self._parent.authenticate(username, password)
        if sessionID == '' and password != '':
            # non authentifié par un parent, on essaye de le faire localement
            try:
                if self._data_proxy.authenticate(username, password):
                    sessionID = self.user_sessions.add_session(username)
                    if sessionID != "":
                        log.msg("Session créée (%s)" % username)
                    return sessionID
            except ldap.SERVER_DOWN:
                log.msg("! Erreur d'accès au serveur LDAP !")
                return ''
            log.msg("! Echec de l'authentification : %s !" % username)
            return ''
        else:
            if sessionID != '':
                log.msg('Session délivrée par le serveur parent (%s)' % username)
            return sessionID

    def get_app_ticket(self, session, appurl):
        # si cette session utilisateur est valide, on crée un ticket pour l'application
        appSessionID = ''
        if self.validate_session(session):
            appurl = get_service_from_url(appurl)
            id_filter = self._check_filter(appurl)
            log.msg("adding session for service  : %s" % appurl)
            appSessionID = self.app_sessions.add_session([appurl,session,id_filter])
            if id_filter != '':
                filter_msg = ' (filtre de données : %s)' % id_filter
            else:
                filter_msg = ''
            log.msg("Session autorisée pour l'application %s%s" % (appurl, filter_msg))
            return appSessionID
        else:
            log.msg("session utilisateur non reconnue !")
        log.msg("! Echec de création de session pour l'application %s :" % appurl)
        return ''
    
    def validate_session(self, session_id):
        """vérifie un session_id existant"""
        if self.user_sessions.validate_session(session_id):
            return True
        if self._parent is not None:
            valid_session = self._parent.validate_session(session_id)
            if valid_session:
                log.msg('Session utilisateur validée par le serveur parent')
            return valid_session
        log.msg("! Session invalide : %s !" % session_id)
        return False

    def verify_session_id(self, session_id):
        """vérifie un session_id"""
        if self.user_sessions.validate_session(session_id):
            return True,session_id
            #try:
            #    if self.user_sessions.get_session_info(session_id) == username:
            #        return True,session_id
            #except:
            #    pass
        else:
            if self._parent is not None:
                return self._parent.verify_session_id(session_id)
        log.msg("! session utilisateur non reconnue : %s !" % session_id)
        return False,session_id


    def get_user_info(self, session_id):
        """méthode de convenance pour récupérer l'utilisateur"""
        return self.user_sessions.get_session_info(session_id)
    

    def logout(self, session_id):
        """vérifie un session_id"""
        if self.user_sessions.validate_session(session_id):
            self.user_sessions.end_session(session_id)
            log.msg('Fin de la session %s' % session_id)
            return 'ok'
        else:
            if self._parent is not None:
                return self._parent.logout(session_id)
            else:
                return ''

    def verify_app_ticket(self, app_ticket, appurl):
        """vérifie le ticket fourni par une application"""
        if self.app_sessions.validate_session(app_ticket):
            try:
                if self.app_sessions.get_session_info(app_ticket)[0] == get_service_from_url(appurl):
                    return True, app_ticket
            except:
                pass
        else:
            if self._parent is not None:
                return self._parent.verify_app_ticket(app_ticket, appurl)
        log.msg("! Echec de la vérification de session pour l'application %s !" % appurl)
        return False, app_ticket

    def get_user_details(self, app_ticket, appurl, sections=False):
        """renvoie les informations sur l'utilisateur
        section : si True, on renvoie aussi une description de l'organisation des données
        """
        log.msg('--- validating session for %s' % (appurl))
        if self.app_sessions.validate_session(app_ticket):
            log.msg('--- session ok')
            code = 'INVALID_SERVICE'
            detail = """Ce ticket n'appartient pas au service %s""" % appurl
            try:
                stored_url, session_id, id_filter = self.app_sessions.get_session_info(app_ticket)
                if stored_url == get_service_from_url(appurl):
                    log.msg('--- appurl verified')
                    return True, self._get_user_details(session_id, id_filter, sections)
            except:
                log.msg("! Lecture des données utilisateur : session invalide !")
                # la session n'existe pas
                pass
        else:
            code = 'INVALID_TICKET'
            detail = """Le ticket %s n'est pas reconnu par le serveur SSO""" % app_ticket
            # session inconnue sur ce serveur, on essaie sur le serveur parent
            if self._parent is not None:
                # récupération des infos sur le serveur parent
                return self._parent.get_user_details(app_ticket, appurl)
        log.msg("! Lecture des données utilisateur refusé pour l'application %s !" % appurl)
        if sections:
            return False, ({'code':code,'detail':detail},"")
        return False, {'code':code,'detail':detail}

    def _check_filter(self, appurl):
        """vérifie si un filtre est disponible pour l'url"""
        url = urlparse.urlparse(appurl)
        if url.scheme == 'https':
            store = self.secure_apps
        else:
            store = self.apps
        # recherche d'une application avec le même port/chemin
        if url.port in store:
            apps = store[url.port]
        else:
            # si pas de port correspondant, on regarde dans les applis sans port spécifique
            apps = store['all']
        if url.path in apps:
            scheme, addr, typeaddr, id_filter = apps[url.path]
            if typeaddr == 'ip':
                addr_ok = check_hostname_by_ip(url, addr)
            else:
                addr_ok = check_hostname_by_domain(url, addr)
            if addr_ok:
                # on a une appli correspondant à l'url, on renvoie le filtre associé
                log.msg("Filtre %s appliqué pour l'url %s" % (id_filter, appurl))
                return id_filter
        return ""
        
    
    def _filter_data(self, infos, id_filter, sections):
        """filtrage en fonction de l'application si des filtres sont définis"""
        data = {}
        if id_filter in self.filters:
            for section, vars in self.filters[id_filter].items():
                for libelle_cas, nom_val in vars.items():
                    data[nom_val] = infos.get(nom_val, '')
            filter_data = self.filters[id_filter]
        else:
            # FIXME : définir les données à renvoyer par défaut !
            data.update(infos)
            filter_data = ""
        if sections:
            return data, filter_data
        else:
            return data

    def _get_user_details(self, session_id, id_filter, sections):
        """Renvoie les informations connues de l'utilisateur lié à <session_id>"""
        # on récupère l'id de l'utilisateur
        user_id = self.user_sessions.get_session_info(session_id)
        infos = self._data_proxy.get_data(user_id)
        # lancement des fonctions de calcul de valeurs supplémentaires
        for varname, calc_func in user_infos.dict_infos.items():
            try:
                infos[varname] = calc_func(infos)
            except exception, e:
                log.msg("""erreur lors du calcul du champ '%s' :  %s""" % (varname, str(e)))

        infos = self._filter_data(infos, id_filter, sections)
        log.msg('--- user data sent')
        return infos


def run(*args):
    global PARENT_URL
    # intialisation du serveur
    from eolesso.dataproxy import LDAPProxy
    from twisted.internet import reactor
    from twisted.web2 import server
    from eolesso.libsecure import ServerContextFactory
    proxy = LDAPProxy(LDAP_SERVER,static_data={'rne':[RNE],'nom_etab':[ETABLISSEMENT]})
    # lancement du serveur
    if PARENT_URL == '':
        PARENT_URL = None
    reactor.listenSSL(int(AUTH_SERVER_PORT), server.Site(SSOSessionManager(proxy,PARENT_URL)),ServerContextFactory(CERTFILE))
    print "\nServeur initialisé (port %s)\n" % AUTH_SERVER_PORT
    reactor.run()


if __name__ == '__main__':
    run()
